/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * linux/arch/mips/kernel/signal.c
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 * Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999  Ralf Baechle
 * Copyright (C) 1999  Silicon Graphics, Inc.
 * Copyright (C) 2000  Sony Computer Entertainment Inc.
 *
 */
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/wait.h>
#include <linux/ptrace.h>
#include <linux/unistd.h>

#include <asm/asm.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/stackframe.h>
#include <asm/uaccess.h>
#include <asm/ucontext.h>

#define DEBUG_SIG 0

#define _BLOCKABLE (~(sigmask(SIGKILL) | sigmask(SIGSTOP)))

asmlinkage void (*lazy_fpu_switch)(struct task_struct *p);
asmlinkage void (*save_fp) (struct task_struct * p);
asmlinkage int sys_wait4(pid_t pid, unsigned long *stat_addr,
                         int options, unsigned long *ru);
asmlinkage int do_signal(sigset_t *oldset, struct pt_regs *regs);
extern asmlinkage void (*save_fp_context)(struct sigcontext *sc);
extern asmlinkage void (*restore_fp_context)(struct sigcontext *sc);

#ifdef CONFIG_CONTEXT_R5900
inline static
int __copy_from_u128(void *to_kern,  void *from_u)
{

	u128 tmp;
	int err = -EFAULT;
__asm__ __volatile__( 
"1:\n\t"
	"lq	%1, (%2);\n\t"
	"sq	%1, (%3);\n\t"
	"move	%0, $0;\n\t"
        ".section	__ex_table,\"a\";\n\t"
        ".word	1b, %4;\n\t"
        ".previous;\n\t"
        :"=r" (err), "=r" (tmp)
        :"r"(from_u), "r"(to_kern), "i" (&&out));
out:
	return err;
}

inline static
int __copy_to_u128(void *from_kern,  void *to_u)
{

	u128 tmp;
	int err = -EFAULT;
__asm__ __volatile__( 
	"lq	%1, (%2);\n\t"
"1:\n\t"
	"sq	%1, (%3);\n\t"
	"move	%0, $0;\n\t"
        ".section	__ex_table,\"a\";\n\t"
        ".word	1b, %4;\n\t"
        ".previous;\n\t"
        :"=r" (err), "=r" (tmp)
        :"r"(from_kern), "r"(to_u), "i" (&&out));
out:
	return err;
}
#endif /* CONFIG_CONTEXT_R5900 */

int copy_siginfo_to_user(siginfo_t *to, siginfo_t *from)
{
	if (!access_ok (VERIFY_WRITE, to, sizeof(siginfo_t)))
		return -EFAULT;
	if (from->si_code < 0)
		return __copy_to_user(to, from, sizeof(siginfo_t));
	else {
		int err;

		/* If you change siginfo_t structure, please be sure
		   this code is fixed accordingly.
		   It should never copy any pad contained in the structure
		   to avoid security leaks, but must copy the generic
		   3 ints plus the relevant union member.  */
		err = __put_user(from->si_signo, &to->si_signo);
		err |= __put_user(from->si_errno, &to->si_errno);
		err |= __put_user((short)from->si_code, &to->si_code);
		/* First 32bits of unions are always present.  */
		err |= __put_user(from->si_pid, &to->si_pid);
		switch (from->si_code >> 16) {
		case __SI_FAULT >> 16:
			break;
		case __SI_CHLD >> 16:
			err |= __put_user(from->si_utime, &to->si_utime);
			err |= __put_user(from->si_stime, &to->si_stime);
			err |= __put_user(from->si_status, &to->si_status);
		default:
			err |= __put_user(from->si_uid, &to->si_uid);
			break;
		/* case __SI_RT: This is not generated by the kernel as of now.  */
		}
		return err;
	}
}

/*
 * Atomically swap in the new signal mask, and wait for a signal.
 */
asmlinkage inline int
sys_sigsuspend(struct pt_regs regs)
{
	sigset_t *uset, saveset, newset;

	save_static(&regs);
	uset = (sigset_t *) get_gpreg(&regs, 4);
	if (copy_from_user(&newset, uset, sizeof(sigset_t)))
		return -EFAULT;
	sigdelsetmask(&newset, ~_BLOCKABLE);

	spin_lock_irq(&current->sigmask_lock);
	saveset = current->blocked;
	current->blocked = newset;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	set_gpreg(&regs, 2, EINTR);
	set_gpreg(&regs, 7, 1);
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		if (do_signal(&saveset, &regs))
			return -EINTR;
	}
}

asmlinkage int
sys_rt_sigsuspend(struct pt_regs regs)
{
	sigset_t *uset, saveset, newset;

	save_static(&regs);
	uset = (sigset_t *) get_gpreg(&regs, 4);
	if (copy_from_user(&newset, uset, sizeof(sigset_t)))
		return -EFAULT;
	sigdelsetmask(&newset, ~_BLOCKABLE);

	spin_lock_irq(&current->sigmask_lock);
	saveset = current->blocked;
	current->blocked = newset;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	set_gpreg(&regs, 2, EINTR);
	set_gpreg(&regs, 7, 1);
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		if (do_signal(&saveset, &regs))
			return -EINTR;
	}
}

asmlinkage int 
sys_sigaction(int sig, const struct sigaction *act,
	      struct sigaction *oact)
{
	struct k_sigaction new_ka, old_ka;
	int ret;

	if (act) {
		if (verify_area(VERIFY_READ, act, sizeof(*act)) ||
		    __get_user(new_ka.sa.sa_handler, &act->sa_handler) ||
		    __get_user(new_ka.sa.sa_flags, &act->sa_flags))
			return -EFAULT;

		copy_from_user( &new_ka.sa.sa_mask, &act->sa_mask, 
				sizeof(sigset_t));
		new_ka.ka_restorer = NULL;
	}

	ret = do_sigaction(sig, act ? &new_ka : NULL, oact ? &old_ka : NULL);

	if (!ret && oact) {
		if (verify_area(VERIFY_WRITE, oact, sizeof(*oact)) ||
		    __put_user(old_ka.sa.sa_handler, &oact->sa_handler) ||
		    __put_user(old_ka.sa.sa_flags, &oact->sa_flags))
			return -EFAULT;
		copy_to_user( &oact->sa_mask, &old_ka.sa.sa_mask, 
		               sizeof(sigset_t));
	}

	return ret;
}

asmlinkage int
sys_sigaltstack(const stack_t *uss, stack_t *uoss)
{
	struct pt_regs *regs = (struct pt_regs *) &uss;

	return do_sigaltstack(uss, uoss, get_gpreg(regs, 29));
}

/*
 * To do: this entire function should be accessed over a function pointer
 * such that we can handle stack frames for different ABIs.
 */

asmlinkage void
restore_sigcontext(struct pt_regs *regs, struct sigcontext *context)
{
	long long reg;
	int owned_fp;

	__get_user(regs->cp0_epc, &context->sc_pc);

#ifdef CONFIG_CONTEXT_R5900
	__copy_from_u128(&regs->hi, &context->sc_mdhi);
	__copy_from_u128(&regs->lo, &context->sc_mdlo);
	__get_user(reg, &context->sc_sa);
	regs->sa = (int) reg;
#else
	__get_user(reg, &context->sc_mdhi);
	regs->hi = (int) reg;
	__get_user(reg, &context->sc_mdlo);
	regs->lo = (int) reg;
#endif

#ifdef CONFIG_CONTEXT_R5900
#define restore_gp_reg(i) __copy_from_u128(&regs->regs[i], &context->sc_regs[i])
#else
#define restore_gp_reg(i) __get_user(reg, &context->sc_regs[i]); \
			  regs->regs[i] = (int) reg;
#endif
	restore_gp_reg( 1); restore_gp_reg( 2); restore_gp_reg( 3);
	restore_gp_reg( 4); restore_gp_reg( 5); restore_gp_reg( 6);
	restore_gp_reg( 7); restore_gp_reg( 8); restore_gp_reg( 9);
	restore_gp_reg(10); restore_gp_reg(11); restore_gp_reg(12);
	restore_gp_reg(13); restore_gp_reg(14); restore_gp_reg(15);
	restore_gp_reg(16); restore_gp_reg(17); restore_gp_reg(18);
	restore_gp_reg(19); restore_gp_reg(20); restore_gp_reg(21);
	restore_gp_reg(22); restore_gp_reg(23); restore_gp_reg(24);
	restore_gp_reg(25); restore_gp_reg(26); restore_gp_reg(27);
	restore_gp_reg(28); restore_gp_reg(29); restore_gp_reg(30);
	restore_gp_reg(31);
#undef restore_gp_reg

	/* FP depends on what FPU in what mode we have.  */
	__get_user(owned_fp, &context->sc_ownedfp);
	if (owned_fp) {
		if (last_task_used_math && (last_task_used_math!=current)) {
			struct pt_regs *last_regs;

			/* Save LAST_TASK_USED_MATH fpu context */
			set_cp0_status(ST0_CU1, ST0_CU1);
			save_fp(last_task_used_math);
			set_cp0_status(ST0_CU1, 0);

			/* Make LAST_TASK_USED_MATH lose fpu */
			last_regs = (struct pt_regs *)
				((unsigned long) last_task_used_math +
				KERNEL_STACK_SIZE - 32
				- sizeof(struct pt_regs));
			last_regs->cp0_status &= ~ST0_CU1;

		}
		restore_fp_context(context);
		last_task_used_math = current;
		current->used_math = 1;
		regs->cp0_status |= ST0_CU1;
	} else {
		/* Mark CURRENT's fpu context to be initilized */
		current->used_math = 0;
		if ( last_task_used_math == current ) {
			/* Make CURRENT lose fpu */
			last_task_used_math = NULL;
			regs->cp0_status &= ~ST0_CU1;
		}
	}
}

/*
 * The structure sc describes the stackframe on the userstack.  The frames
 * are identical for normal and realtime signal stackframes.
 */

#if DEBUG_SIG
#define SIGFRAME_GUARD_SIZE 12
#define SIGFRAME_CHECK_GUARD
#define SIGFRAME_GUARD_MAGIC 0xf571175f
#endif

struct sigframe {
	unsigned long	sf_ass[4];	/* argument save space for o32 */
#ifdef SIGFRAME_GUARD_SIZE
	unsigned long   sf_guard[SIGFRAME_GUARD_SIZE];
#endif
#ifdef CONFIG_CPU_R5900
	unsigned int	sf_code[4];	/* signal trampoline */
#else
	unsigned int	sf_code[2];	/* signal trampoline */
#endif
	siginfo_t       sf_info;
	struct ucontext sf_uc;
};

asmlinkage int sys_sigreturn(struct pt_regs regs)
{
	struct sigframe *frame;
	struct sigcontext *context;
	sigset_t blocked;
#ifdef SIGFRAME_CHECK_GUARD
	int i;
#endif

	frame = (struct sigframe *) get_gpreg(&regs, 29);
	context = &frame->sf_uc.uc_mcontext;

#ifdef SIGFRAME_CHECK_GUARD
	for (i=0; i < SIGFRAME_GUARD_SIZE; i ++) 
	    if (frame->sf_guard[i] != SIGFRAME_GUARD_MAGIC) 
		printk("%s:SIG handler may break stack:%d:[%lx]:%lx\n",
		    current->comm, i,
		    (unsigned long)&frame->sf_guard[i], 
		    (unsigned long)frame->sf_guard[i]);
#endif
	if (!access_ok(VERIFY_READ, context, sizeof(struct sigcontext)) ||
	    ((unsigned long)context & (SZREG - 1)))
		goto badframe;

	if (__get_user(blocked.sig[0], &context->sc_sigset[0]) ||
	    __get_user(blocked.sig[1], &context->sc_sigset[1]) ||
	    __get_user(blocked.sig[2], &context->sc_sigset[2]) ||
	    __get_user(blocked.sig[3], &context->sc_sigset[3]))
		goto badframe;

	sigdelsetmask(&blocked, ~_BLOCKABLE);
	spin_lock_irq(&current->sigmask_lock);
	current->blocked = blocked;
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	restore_sigcontext(&regs, context);

	/*
	 * Don't let your children do this ...
	 */
	__asm__ __volatile__(
		"move\t$29,%0\n\t"
		"j\tret_from_sys_call"
		:/* no outputs */
		:"r" (&regs));
	/* Unreached */

badframe:
	lock_kernel();
	do_exit(SIGSEGV);
	unlock_kernel();
}

/* same as sys_sigreturn for now */
asmlinkage int sys_rt_sigreturn(struct pt_regs regs)
{
	return -ENOSYS;
}


/*
 * Set up the return code ...
 *
 *         .set    noreorder
 *         li      v0,__NR_sigreturn
 *         syscall
 *         .set    reorder
 */
static void inline
setup_trampoline(unsigned int *code)
{
	__put_user(0x24020000 + __NR_sigreturn, code + 0);
	__put_user(0x0000000c                 , code + 1);
#ifdef CONFIG_CPU_R5900
	__put_user(0x0000040f		      , code + 2);	/* sync.p */
	__put_user(0			      , code + 3);	/* nop    */
#endif

	/*
	 * Flush caches so that the instructions will be correctly executed.
	 */
	flush_cache_sigtramp((unsigned long) code);
}

static void inline
setup_sigcontext(struct pt_regs *regs, struct sigcontext *sc, sigset_t *set)
{
	int owned_fp;

	__put_user(regs->cp0_epc, &sc->sc_pc);
	__put_user(regs->cp0_status, &sc->sc_status);	/* Status register */

#ifdef CONFIG_CONTEXT_R5900
#define save_gp_reg(i) __copy_to_u128(&regs->regs[(i)], &sc->sc_regs[(i)])
	{ u128 zero=0;
	__copy_to_u128(&zero, &sc->sc_regs[0]); 
	}
	save_gp_reg(1); save_gp_reg(2);
#else
#define save_gp_reg(i) __put_user(regs->regs[(i)], &sc->sc_regs[(i)])
	__put_user(0, &sc->sc_regs[0]); save_gp_reg(1); save_gp_reg(2);
#endif
	save_gp_reg(3); save_gp_reg(4); save_gp_reg(5); save_gp_reg(6);
	save_gp_reg(7); save_gp_reg(8); save_gp_reg(9); save_gp_reg(10);
	save_gp_reg(11); save_gp_reg(12); save_gp_reg(13); save_gp_reg(14);
	save_gp_reg(15); save_gp_reg(16); save_gp_reg(17); save_gp_reg(18);
	save_gp_reg(19); save_gp_reg(20); save_gp_reg(21); save_gp_reg(22);
	save_gp_reg(23); save_gp_reg(24); save_gp_reg(25); save_gp_reg(26);
	save_gp_reg(27); save_gp_reg(28); save_gp_reg(29); save_gp_reg(30);
	save_gp_reg(31);
#undef save_gp_reg

#ifdef CONFIG_CONTEXT_R5900
	__copy_to_u128(&regs->hi, &sc->sc_mdhi);
	__copy_to_u128(&regs->lo, &sc->sc_mdlo);
	__put_user(regs->sa, &sc->sc_sa);
#else
	__put_user(regs->hi, &sc->sc_mdhi);
	__put_user(regs->lo, &sc->sc_mdlo);
#endif
	__put_user(regs->cp0_cause, &sc->sc_cause);

	/* Load CURRENT's fpu context, if CURRENT have used fpu. */
	if ( current->used_math && (current != last_task_used_math)) {
		lazy_fpu_switch(last_task_used_math);
		last_task_used_math = current;
	}

	owned_fp = (current == last_task_used_math);
	__put_user(owned_fp, &sc->sc_ownedfp);

	if (owned_fp) {	/* fp is active.  */
		set_cp0_status(ST0_CU1, ST0_CU1);
		save_fp_context(sc);		/* CPU-dependent */

		/* Make CURRENT lose fpu */
		save_fp(current);
		last_task_used_math = NULL;
		regs->cp0_status &= ~ST0_CU1;

		set_cp0_status(ST0_CU1, 0);
	}

	__put_user(set->sig[0], &sc->sc_sigset[0]);
	__put_user(set->sig[1], &sc->sc_sigset[1]);
	__put_user(set->sig[2], &sc->sc_sigset[2]);
	__put_user(set->sig[3], &sc->sc_sigset[3]);
}

/*
 * Determine which stack to use..
 */
static inline void *
get_sigframe(struct k_sigaction *ka, struct pt_regs * regs, size_t frame_size)
{
  unsigned long sp;

  sp = get_gpreg(regs, 29);
  /* This is the X/Open sanctioned signal stack switching.  */
  if (ka->sa.sa_flags & SA_ONSTACK) {
    if (! on_sig_stack(sp))
      sp = current->sas_ss_sp + current->sas_ss_size;
  }
  return (void *)((sp - frame_size) & ALMASK);
}

static void inline
setup_frame(struct k_sigaction * ka, struct pt_regs *regs,
            int signr, sigset_t *oldmask, siginfo_t *info)
{
	struct sigframe *frame;
	unsigned int dummy;
#ifdef SIGFRAME_CHECK_GUARD
	int i;
#endif

	frame = get_sigframe(ka, regs, sizeof(*frame));

	if (verify_area(VERIFY_WRITE, frame, sizeof (*frame)))
		goto segv_and_exit;

	/* make sure accessable stack is there */
	if (__get_user(dummy, &frame->sf_ass[0]))
		goto segv_and_exit;

	setup_trampoline(frame->sf_code);
	setup_sigcontext(regs, &frame->sf_uc.uc_mcontext, oldmask);

#ifdef SIGFRAME_CHECK_GUARD
	for (i=0; i < SIGFRAME_GUARD_SIZE; i ++) {
	  frame->sf_guard[i] = (unsigned long)SIGFRAME_GUARD_MAGIC;
	}
#endif

	/*
	 * Arguments to signal handler:
	 * for SA_SIGINFO specified
	 *   a0 = signal number
	 *   a1 = pointer to struct siginfo
	 *   a2 = pointer to struct ucontext
	 * otherwise
	 *   a0 = signal number
	 *   a1 = cause
	 *   a2 = pointer to struct sigcontext
	 *
	 * $25 and c0_epc point to the signal handler, $29 points to the
	 * struct sigframe.
	 */

	set_gpreg(regs, 4, signr);			/* arguments */
	if (info != NULL)  {
		int err = 0;

		/* Create siginfo.  */
		err |= copy_siginfo_to_user(&frame->sf_info, info);

		/* Create the ucontext.  */
		err |= __put_user(0, &frame->sf_uc.uc_flags);
		err |= __put_user(0, &frame->sf_uc.uc_link);
		err |= __put_user((void *)current->sas_ss_sp,
				  &frame->sf_uc.uc_stack.ss_sp);
		err |= __put_user(sas_ss_flags(regs->regs[29]),
				  &frame->sf_uc.uc_stack.ss_flags);
		err |= __put_user(current->sas_ss_size,
				  &frame->sf_uc.uc_stack.ss_size);
		err |= __copy_to_user(&frame->sf_uc.uc_sigmask, oldmask, 
					sizeof(*oldmask));

		if (err)
			goto segv_and_exit;

		/* ptr to siginfo */
		set_gpreg(regs, 5, (unsigned long)&frame->sf_info);
		/* ptr to ucontext */
		set_gpreg(regs, 6, (unsigned long)&frame->sf_uc);

	} else {
		/* cause */
		set_gpreg(regs, 5, (&frame->sf_uc.uc_mcontext)->sc_cause);
		/* ptr to sigcontext */
		set_gpreg(regs, 6, (unsigned long) &frame->sf_uc.uc_mcontext);
	}

	set_gpreg(regs, 29, (unsigned long)frame); 	/* Stack pointer */
	set_gpreg(regs, 31, (unsigned long) frame->sf_code); /* Return address */
        regs->cp0_epc = (unsigned long) ka->sa.sa_handler; /* handler address */
	set_gpreg(regs, 25, (unsigned long) regs->cp0_epc); /* PIC shit... */

#if DEBUG_SIG
	printk("SIG deliver (%s:%d): sp=0x%p pc=0x%p ra=0x%p\n",
	       current->comm, current->pid, frame, regs->cp0_epc, frame->sf_code);
#endif
        return;

segv_and_exit:
	lock_kernel();
	do_exit(SIGSEGV);
	unlock_kernel();
}


/* ------------------------------------------------------------------------- */

static inline void handle_signal(unsigned long sig, struct k_sigaction *ka,
	siginfo_t *info, sigset_t *oldset, struct pt_regs * regs)
{
	if (ka->sa.sa_flags & SA_SIGINFO)
		setup_frame(ka, regs, sig, oldset, info);
	else
		setup_frame(ka, regs, sig, oldset, NULL);

	if (ka->sa.sa_flags & SA_ONESHOT)
		ka->sa.sa_handler = SIG_DFL;
	if (!(ka->sa.sa_flags & SA_NODEFER)) {
		spin_lock_irq(&current->sigmask_lock);
		sigorsets(&current->blocked,&current->blocked,&ka->sa.sa_mask);
		sigaddset(&current->blocked,sig);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);
	}
}

static inline void syscall_restart(struct pt_regs *regs, struct k_sigaction *ka)
{
	switch(get_gpreg(regs,0)) {
	case ERESTARTNOHAND:
		set_gpreg(regs,2,EINTR);
		break;
	case ERESTARTSYS:
		if(!(ka->sa.sa_flags & SA_RESTART)) {
			set_gpreg(regs,2,EINTR);
			break;
		}
	/* fallthrough */
	case ERESTARTNOINTR:		/* Userland will reload $v0.  */
		set_gpreg(regs,7,get_gpreg(regs, 26));
		regs->cp0_epc -= 8;
	}

	set_gpreg(regs, 0, 0);		/* Don't deal with this again.  */
}

extern int do_irix_signal(sigset_t *oldset, struct pt_regs *regs);

asmlinkage int do_signal(sigset_t *oldset, struct pt_regs *regs)
{
	struct k_sigaction *ka;
	siginfo_t info;

#ifdef CONFIG_BINFMT_IRIX
	if (current->personality != PER_LINUX)           /* XXX */
		return do_irix_signal(oldset, regs);
#endif

	if (!oldset)
		oldset = &current->blocked;

	for (;;) {
		unsigned long signr;

		spin_lock_irq(&current->sigmask_lock);
		signr = dequeue_signal(&current->blocked, &info);
		spin_unlock_irq(&current->sigmask_lock);

		if (!signr)
			break;

		if ((current->flags & PF_PTRACED) && signr != SIGKILL) {
			/* Let the debugger run.  */
			current->exit_code = signr;
			current->state = TASK_STOPPED;
			notify_parent(current, SIGCHLD);
			schedule();

			/* We're back.  Did the debugger cancel the sig?  */
			if (!(signr = current->exit_code))
				continue;
			current->exit_code = 0;

			/* The debugger continued.  Ignore SIGSTOP.  */
			if (signr == SIGSTOP)
				continue;

			/* Update the siginfo structure.  Is this good?  */
			if (signr != info.si_signo) {
				info.si_signo = signr;
				info.si_errno = 0;
				info.si_code = SI_USER;
				info.si_pid = current->p_pptr->pid;
				info.si_uid = current->p_pptr->uid;
			}

			/* If the (new) signal is now blocked, requeue it.  */
			if (sigismember(&current->blocked, signr)) {
				send_sig_info(signr, &info, current);
				continue;
			}
		}

		ka = &current->sig->action[signr-1];
		if (ka->sa.sa_handler == SIG_IGN) {
			if (signr != SIGCHLD)
				continue;
			/* Check for SIGCHLD: it's special.  */
			while (sys_wait4(-1, NULL, WNOHANG, NULL) > 0)
				/* nothing */;
			continue;
		}

		if (ka->sa.sa_handler == SIG_DFL) {
			int exit_code = signr;

			/* Init gets no signals it doesn't want.  */
			if (current->pid == 1)
				continue;

			switch (signr) {
			case SIGCONT: case SIGCHLD: case SIGWINCH:
				continue;

			case SIGTSTP: case SIGTTIN: case SIGTTOU:
				if (is_orphaned_pgrp(current->pgrp))
					continue;
				/* FALLTHRU */

			case SIGSTOP:
				current->state = TASK_STOPPED;
				current->exit_code = signr;
				if (!(current->p_pptr->sig->action[SIGCHLD-1].sa.sa_flags & SA_NOCLDSTOP))
					notify_parent(current, SIGCHLD);
				schedule();
				continue;

			case SIGQUIT: case SIGILL: case SIGTRAP:
			case SIGABRT: case SIGFPE: case SIGSEGV:
			case SIGBUS:
				if (do_coredump(signr, regs))
					exit_code |= 0x80;
				/* FALLTHRU */
			default:
				lock_kernel();
				sigaddset(&current->signal, signr);
				current->flags |= PF_SIGNALED;
				do_exit(exit_code);
				/* NOTREACHED */
			}
		}

		if (get_gpreg(regs, 0))
			syscall_restart(regs, ka);
		/* Whee!  Actually deliver the signal.  */
		handle_signal(signr, ka, &info, oldset, regs);
		return 1;
	}

	/*
	 * Who's code doesn't conform to the restartable syscall convention
	 * dies here!!!  The li instruction, a single machine instruction,
	 * must directly be followed by the syscall instruction.
	 */
	if (get_gpreg(regs, 0)) {
		if (get_gpreg(regs, 2) == ERESTARTNOHAND ||
		    get_gpreg(regs, 2) == ERESTARTSYS ||
		    get_gpreg(regs, 2) == ERESTARTNOINTR) {
			set_gpreg(regs, 7, get_gpreg(regs, 26));
			regs->cp0_epc -= 8;
		}
	}
	return 0;
}

/*
 * Compatibility syscall.  Can be replaced in libc.
 */
asmlinkage int sys_pause(void)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule();
	return -ERESTARTNOHAND;
}
